#!/usr/bin/ruby

require 'getopts'

class IO

    def readrec(sizelimit = nil)
	len = read(4)
	raise EOFError if len.nil?
	raise 'unexpected eof' if len.size < 4
	len = len.unpack('N').first
	unless sizelimit.nil?
	    raise "record size #{len} > #{sizelimit}" if len >= sizelimit
	end
	raise "record length #{len} too short at #{tell}" if len < 8
	body = read(len - 8)
	len2 = read(4)
	raise 'unexpected eof' if len2.nil?
	len2 = len2.unpack('N').first
	raise "malformatted record #{len} != #{len2}" if len != len2
	body
    end

    def writerec(buf)
	str = buf.to_s
	rechdr = [str.size + 8].pack('N')
	write(rechdr + str + rechdr)
    end

end

class LogFile < File

    def self::open(*args)
	filename, mode = args
	if mode == 'w' and File.exist? filename
	    stat = File::stat(filename)
	    if not stat.writable?
		newmode = stat.mode | 0222
		printf "     chmod %o %s\n", newmode, filename if $VERBOSE
		File::chmod(newmode, filename)
	    end
	end
	super.ungetblock(nil)
    end

    def self::delete(filename)
	print "dele #{filename}\n" if $VERBOSE
	super
    end

    def readblock
	title = gets
	return nil if title.nil?
	raise "malformatted =begin line" unless /^=begin/ =~ title
	info = Hash::new
	title = gets
	raise "malformatted title line #{title}" \
	    unless %r|^= (\w+) ([\w./-]*)| =~ title
	info['kind'] = $1
	path = $2
	info['path'] = path.urldecode
	while line = gets
	    case line
	    when /^=end/
	    	break
	    when /^: *(\w+)/
		key = $1
		line = readline.gsub(/\s*/, '')
		info[key] = Integer(line)
	    when /^\+ *(\w+)/
		key = $1
		buffer = ""
		while tmp = readline
		    buffer += tmp
		    break if /^\s*$/ =~ buffer
		end
		info[key] = buffer.unpack('u').first
	    end
	end
	return info
    end

    def getblock
	if @ungetblock
	    r = @ungetblock
	    @ungetblock = nil
	    r
	else
	    readblock
	end
    end

    def ungetblock h
	@ungetblock = h
	self
    end

end

class String

    def urlencode
	gsub(/[^-\w.\/]/){|c| format('%%%02X', c[0])}
    end

    def urldecode
	gsub(/%../){|c| Integer('0x' + c[1..2]).chr}
    end

end

def filecheck(filename, log, size)
    fp = File::open(filename)
    fp.binmode
    begin
	recN = fp.readrec(size)
	fp.readrec(size)
	ofsI = fp.tell
	recI = fp.readrec(size)
	raise if size != recN.unpack('x96N').first
	fp.seek(size - 4)
	recsiz = fp.read(4).unpack('N').first
	fp.seek(-recsiz, IO::SEEK_END)
	ofsE = fp.tell
	recE = fp.readrec(size)
	print "nusf #{filename}\n" if $VERBOSE
	log.print <<EOF
=begin
= NUSF #{filename.urlencode}
: size
  #{size}
: ofs_indx
  #{ofsI}
: ofs_end
  #{ofsE}
+ NUSD
#{[recN].pack('u')}
+ INDX
#{[recI].pack('u')}
+ END
#{[recE].pack('u')}
=end
EOF
    rescue => e
	print "file #{filename}\n" if $VERBOSE
	p e.to_s if $DEBUG
	log.print <<EOF
=begin
= FILE #{filename.urlencode}
=end
EOF
    end
    fp.close
end

def create_r(dirname, log)
    d = Dir::open(dirname)
    children = (d.collect - ['..', '.']).sort
    d.close
    log.print <<EOF
=begin
= DIR #{dirname.urlencode}
: children
  #{children.size}
=end
EOF
    for fnam in children
    	child = dirname + '/' + fnam
	stat = File::lstat(child)
	if (stat.directory?) then
    	    create_r(child, log)
	elsif (stat.file?) then
	    filecheck(child, log, stat.size)
	else
	    raise "special file #{child}: #{stat}"
	end
    end
end

def nusdtest(filename, info)
    print "nusd #{filename}\n" if $VERBOSE
    size, ofsI, ofsE = info['size'], info['ofs_indx'], info['ofs_end']
    recN, recI, recE = info['NUSD'], info['INDX'], info['END']
    stat = File::stat filename
    print "     size match: #{stat.size == size}\n" if $VERBOSE
    raise "file truncated too much" if stat.size < ofsE
    begin
	fp = File::open(filename, "r+")
    rescue Errno::EACCES
	newmode = stat.mode | 0222
	printf "     chmod %o %s\n", newmode, filename if $VERBOSE
	File::chmod(newmode, filename)
	fp = File::open(filename, "r+")
    end
    fp.binmode

    # check and repair NUSD record
    writeback = fp.tell
    begin
	newrN = fp.readrec
	print "     NUSD match: #{recN == newrN}\n" if $VERBOSE
	writeback = false if (recN == newrN)
    rescue => e
        print "     NUSD reading error: #{e.message}\n" if $VERBOSE
    end
    if writeback
	fp.seek writeback
	fp.writerec(recN)
        print "     NUSD wrote at ofs #{writeback}\n" if $VERBOSE
	fp.seek writeback + recN.size + 8
    end

    # skip CNTL record
    begin
	fp.readrec
    rescue => e
	STDERR.print "Error in reading CNTL (#{fp.tell})"
	raise e
    end

    # check and repair INDX record
    writeback = fp.tell
    print "     INDX offset: #{ofsI == writeback}\n" if $VERBOSE
    begin
	newrI = fp.readrec
	print "     INDX match: #{recI == newrI}\n" if $VERBOSE
	writeback = false if (recI == newrI)
    rescue => e
        print "     INDX reading error: #{e.message}\n" if $VERBOSE
    end
    if writeback
	fp.seek writeback
	fp.writerec(recI)
        print "     INDX wrote at ofs #{writeback}\n" if $VERBOSE
	fp.seek writeback + recI.size + 8
    end

    # check and repair END record
    fp.seek ofsE
    writeback = ofsE
    begin
	newrE = fp.readrec
	print "     END  match: #{recE == newrE}\n" if $VERBOSE
	writeback = false if (recE == newrE)
    rescue => e
        print "     END reading error: #{e.message}\n" if $VERBOSE
    end
    if writeback
	fp.seek writeback
	fp.writerec(recE)
        print "     END wrote at ofs #{writeback}\n" if $VERBOSE
	fp.seek writeback + recE.size + 8
    end
    raise "size mismatch #{fp.tell} != #{size}" if fp.tell != size
    if stat.size > size
	print "     truncation at #{size}\n" if $VERBOSE
	fp.truncate(size)
    end

    fp.close
end

def restore_r(dirname, log)

    d = Dir::open(dirname)
    children = (d.collect - ['..', '.']).sort
    d.close
    check = Hash::new
    for fnam in children
    	child = dirname + '/' + fnam
	check[child] = true
    end

    info = log.getblock
    if info.nil?
	print " not saved: not restored.\n"
	return nil
    end
    print "dir  #{dirname}\n"
    if info['path'] != dirname
	log.ungetblock info
	print " --> log: #{info['path']}\n"
	return nil
    end
    if info['children'] != children.size
	print "    children #{info['children']} -> #{children.size}\n"
    end

    info['children'].times do
	cinfo = log.getblock
    	child = cinfo['path']
	check[child] = false
	if check[child].nil?
	    print "#{cinfo['kind']} #{child} disappeared\n"
	    raise "#{cinfo['kind']} #{child} disappeared"
	end
	stat = File::stat(child)
	case cinfo['kind']
	when 'DIR'
	    raise "dir -> not dir: #{child}" unless stat.directory?
	    log.ungetblock cinfo
	    restore_r(child, log)
	when 'FILE'
	    raise "file -> not file: #{child}" unless stat.file?
	    print "file #{child}\n" if $VERBOSE
	when /^NUSF/
	    raise "file -> not file: #{child}" unless stat.file?
            nusdtest(child, cinfo)
	else
	    raise("bad file info=" + info.dump)
	end
    end

    for k, unchecked in check
	next if not unchecked
	File::delete k
	print "new  #{k} (deleted)\n"
    end
end

def dircheck(dirname, mode, log0 = nil)
    print "#{dirname}:\n"
    unless File::directory? dirname
	if File::symlink? dirname
	    alt = File::readlink dirname
	    if not File::exist?(alt)
		print "skip #{dirname} -> #{alt} (non-existent)\n"
		return
	    end
	end
    	raise "#{dirname} is not a directory"
    end
    case mode
    when :clear
	LogFile::delete(FILENAME) unless log0
    when :restore
	log = log0 ? log0 : LogFile::open(FILENAME, 'r')
	restore_r(dirname, log)
	log.close unless log0
    else
    	log = log0 ? log0 : LogFile::open(FILENAME, 'w')
	create_r(dirname, log)
	log.close unless log0
    end
end

getopts('dqwcra', 'filename:CHECKPOINT')
FILENAME = $OPT_filename
$DEBUG = $OPT_d
$VERBOSE = ($OPT_w or not $OPT_q)
mode = :make
mode = :clear if $OPT_c
mode = :restore if $OPT_r
PWD = Dir::pwd

log = $OPT_a ? LogFile::open(FILENAME, (mode == :make ? 'w' : 'r')) : nil
for dir in ARGV
    dircheck(dir, mode, log)
    dircheck(dir, :clear, log) if $OPT_c and $OPT_r
end
log.close if log
LogFile::delete FILENAME if log and mode == :clear
