/*
=begin
= class NWPL::NwpTimeSpan
Class of time interval.
It contains integer counter of minutes and pentads internally.
= class NWPL::NwpTime
Class of time.
Internally it is represented with minutes from 00:00 UTC 1 Jan 1801
(in Japanese, KANSEI 12 NEN 11 GATU 17 NITI ASA TATU NO TOKI).
=end
*/

#include <ruby.h>
#include <string.h>
#include <ctype.h>
#include <time.h>
#include "nwplraw.h"

#define DEBUG(fmt)
/* #define DEBUG(fmt)	printf fmt */

#define strieq(a, b) (strcasecmp((a), (b)) == 0)

#define MIN_PER_HOUR 60
#define MIN_PER_DAY (60 * 24)
#define MIN_PER_WEEK (60 * 24 * 7)
#define PEN_PER_YEAR (6 * 12)

VALUE mNwpl;
VALUE nwpl_cNwpTimeSpan;
VALUE nwpl_cNwpTime;

struct NwpTimeSpan {
    long	min;
    long	pentad;
};

struct NwpTime {
    long	min;
    struct tm   st;
};

static inline VALUE
LookupModule(const char *modname)
{
    VALUE idx, nwplname;
    nwplname = rb_str_new2(modname);
    idx = rb_funcall(rb_cObject, rb_intern("const_defined?"), 1, nwplname);
    if (idx == Qtrue) {
	return rb_funcall(rb_cObject, rb_intern("const_get"), 1, nwplname);
    } else {
        return rb_define_module(modname);
    }
}

static void
nwpl_tspan_free(struct NwpTimeSpan *timespan_ptr)
{
    xfree(timespan_ptr);
}

static inline VALUE
new_tspan(long min, long pentad, VALUE klass)
{
    struct NwpTimeSpan *tspan;
    tspan = ALLOC(struct NwpTimeSpan);
    tspan->min = min;
    tspan->pentad = pentad;
    return Data_Wrap_Struct(klass, 0, nwpl_tspan_free, tspan);
}

#define NEW_TSPAN new_tspan

VALUE
nwpl_tspan_new(int argc, VALUE *argv, VALUE klass)
{
    VALUE factor;
    const char *c_units;

    if (argc < 1 || argc > 2) {
	rb_raise(rb_eArgError, "usage: new(factor, [units])");
    }
    factor = argv[0];
    switch (TYPE(factor)) {
    case T_ARRAY:
	if (RARRAY(factor)->len == 2) {
	    return NEW_TSPAN(NUM2LONG(RARRAY(factor)->ptr[0]),
		 NUM2LONG(RARRAY(factor)->ptr[1]), klass);
	} else if (RARRAY(factor)->len == 1) {
            factor = RARRAY(factor)->ptr[0];
	} else {
	    rb_raise(rb_eArgError,
		"if the 1st arg is Array, they must have elements");
        }
    }
    if (argc == 2) {
	c_units = STR2CSTR(argv[1]);
    } else {
	c_units = "MIN";
    }
    if (strieq(c_units, "MIN") || strieq(c_units, "MINUTE")) {
	return NEW_TSPAN(NUM2LONG(factor), 0, klass);
    } else if (strieq(c_units, "HOUR")) {
	return NEW_TSPAN(NUM2LONG(factor) * MIN_PER_HOUR, 0, klass);
    } else if (strieq(c_units, "DAY")) {
	return NEW_TSPAN(NUM2LONG(factor) * MIN_PER_DAY, 0, klass);
    } else if (strieq(c_units, "WEEK")) {
	return NEW_TSPAN(NUM2LONG(factor) * MIN_PER_WEEK, 0, klass);
    } else if (strieq(c_units, "PENT")) {
	return NEW_TSPAN(0, NUM2LONG(factor), klass);
    } else if (strieq(c_units, "MON") || strieq(c_units, "MONTH")) {
	return NEW_TSPAN(0, NUM2LONG(factor) * 6, klass);
    } else if (strieq(c_units, "YEAR")) {
	return NEW_TSPAN(0, NUM2LONG(factor) * PEN_PER_YEAR, klass);
    } else {
        rb_raise(rb_eArgError, "invalid time units <%s>", c_units);
    }
}

VALUE
nwpl_tspan_min(VALUE obj)
{
    struct NwpTimeSpan *tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    return INT2NUM(tspan->min);
}

VALUE
nwpl_tspan_pentad(VALUE obj)
{
    struct NwpTimeSpan *tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    return INT2NUM(tspan->pentad);
}

VALUE
nwpl_tspan_mon(VALUE obj)
{
    struct NwpTimeSpan *tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    if (tspan->pentad % 6 == 0) {
        return INT2NUM(tspan->pentad / 6);
    } else {
        return rb_float_new(tspan->pentad / 6.0);
    }
}

VALUE
nwpl_tspan_to_s(VALUE obj)
{
    struct NwpTimeSpan	*tspan;
    char	buf[256];
    int		len = 0, ofs;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    if (tspan->pentad == 0) {
        len = 0;
    } else if (tspan->pentad % 72 == 0) {
        len = snprintf(buf, 64, "(%ld YEAR + ", tspan->pentad / 72);
    } else if (tspan->pentad % 6 == 0) {
        len = snprintf(buf, 64, "(%ld MON + ", tspan->pentad / 6);
    } else {
        len = snprintf(buf, 64, "(%ld PENT + ", tspan->pentad);
    }
    ofs = len;
    if (tspan->min == 0 && len > 4) {
        return rb_str_new(buf + 1, len - 4);
    } else if (tspan->min % MIN_PER_WEEK == 0) {
        len += snprintf(buf + len, 64, "%ld WEEK", tspan->min / MIN_PER_WEEK);
    } else if (tspan->min % MIN_PER_DAY == 0) {
        len += snprintf(buf + len, 64, "%ld DAY", tspan->min / MIN_PER_DAY);
    } else if (tspan->min % MIN_PER_HOUR == 0) {
        len += snprintf(buf + len, 64, "%ld HOUR", tspan->min / MIN_PER_HOUR);
    } else {
        len += snprintf(buf + len, 64, "%ld MIN", tspan->min);
    }
    if (ofs) {
        strcpy(buf + len, ")");
	len++;
    }
    return rb_str_new(buf, len);
}

VALUE
nwpl_tspan_mul(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
        double d;
        long ell;
        case T_FLOAT:
            d = NUM2DBL(other);
	    return NEW_TSPAN(tspan->min * d, tspan->pentad * d, CLASS_OF(obj));
        case T_FIXNUM:
        case T_BIGNUM:
        default:
            ell = NUM2LONG(other);
	    return NEW_TSPAN(tspan->min * ell, tspan->pentad * ell,
		CLASS_OF(obj));
    }
}

VALUE
nwpl_tspan_div(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
        double d;
        long ell;
        case T_FLOAT:
            d = NUM2DBL(other);
	    return NEW_TSPAN(tspan->min / d, tspan->pentad / d, CLASS_OF(obj));
        case T_FIXNUM:
        case T_BIGNUM:
        default:
            ell = NUM2LONG(other);
	    return NEW_TSPAN(tspan->min / ell, tspan->pentad / ell,
		CLASS_OF(obj));
    }
}

VALUE
nwpl_tspan_neg(VALUE obj)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    return NEW_TSPAN(-tspan->min, -tspan->pentad, CLASS_OF(obj));
}

static void
nwpl_time_free(struct NwpTime *time_ptr)
{
    xfree(time_ptr);
}

static inline VALUE
new_time(long min, VALUE klass)
{
    struct NwpTime *timep;
    timep = ALLOC(struct NwpTime);
    timep->min = min;
    timep->st.tm_isdst = -1;
    return Data_Wrap_Struct(klass, 0, nwpl_time_free, timep);
}

#define NEW_TIME new_time

static void
prepare_tm_essential(struct NwpTime *timep)
{
    if (timep->st.tm_isdst > -1)
	return;
    nwpl_seq2tm(timep->min, &(timep->st));
}

static void
prepare_tm_optional(struct NwpTime *timep)
{
    if (timep->st.tm_isdst != -1)
	return;
    nwpl_seq2tm(timep->min, &(timep->st));
}

VALUE
nwpl_time_new(int argc, VALUE *argv, VALUE klass)
{
    VALUE spec, t;

    if (argc < 1 || argc > 2) {
	rb_raise(rb_eArgError, "usage: new(spec, [quietp])");
    }
    spec = argv[0];
    switch (TYPE(spec)) {
	long	t;
	int	alertp;
        case T_STRING:
	    t = nwpl_str2mseq(RSTRING(spec)->ptr);
	    alertp = (argc < 2) || NIL_P(argv[1]) || (argv[1] == Qfalse);
            if ((t == -1) && alertp) {
                rb_raise(rb_eArgError, "bad format (%s)", RSTRING(spec)->ptr);
	    }
	    return NEW_TIME(t, klass);
	case T_DATA:
	    goto data;
	case T_BIGNUM:
	case T_FIXNUM:
	case T_FLOAT:
	default:
	    return NEW_TIME(NUM2LONG(spec), klass);
    }

data:
    if (CLASS_OF(spec) == nwpl_cNwpTime) {
	t = rb_funcall(spec, rb_intern("to_i"), 0);
	return NEW_TIME(NUM2LONG(t), klass);
    } else if (CLASS_OF(spec) == rb_cTime) {
	unsigned long	s;
	t = rb_funcall(spec, rb_intern("to_i"), 0);
	s = NUM2ULONG(t) / MIN_PER_HOUR + UNIX_ERA;
	return NEW_TIME(s, klass);
    } else {
	return NEW_TIME(NUM2LONG(spec), klass);
    }
}

VALUE
nwpl_time_gettime(int argc, VALUE *argv, VALUE klass)
{
    struct NwpTime	*timep;
    const char		*fnam;
    FILE		*input;
    int			r;

    if (argc > 1) {
	rb_raise(rb_eArgError, "usage: gettime(filename = \"TIMECARD\")");
    } else if (argc == 1) {
	fnam = STR2CSTR(argv[0]);
    } else {
	fnam = "TIMECARD";
    }
    timep = ALLOC(struct NwpTime);
    timep->st.tm_isdst = -2;
    input = fopen(fnam, "rt");
    if (input == NULL)
	rb_raise(rb_eIOError, "cannot open <%s>", fnam);
    r = fscanf(input, "%4u%2u%2u%2u%2u", &(timep->st.tm_year),
	&(timep->st.tm_mon), &(timep->st.tm_mday), &(timep->st.tm_hour),
	&(timep->st.tm_min));
    fclose(input);
    if (r != 5)
	rb_raise(rb_eIOError, "cannot read <%s> (maybe format error)", fnam);
    timep->min = nwpl_ymdhm2seq(timep->st.tm_year, timep->st.tm_mon,
	timep->st.tm_mday, timep->st.tm_hour, timep->st.tm_min);
    timep->st.tm_year -= 1900;
    timep->st.tm_mon--;
    return Data_Wrap_Struct(klass, 0, nwpl_time_free, timep);
}

VALUE
nwpl_time_to_i(VALUE obj)
{
    struct NwpTime	*time;
    Data_Get_Struct(obj, struct NwpTime, time);   
    return INT2NUM(time->min);
}

VALUE
nwpl_time_time(VALUE obj)
{
    struct NwpTime	*time;
    Data_Get_Struct(obj, struct NwpTime, time);   
    return rb_funcall(rb_cTime, rb_intern("at"), 1,
	INT2NUM((time->min - UNIX_ERA) * MIN_PER_HOUR));
}

VALUE
nwpl_time_strftime(VALUE obj, VALUE fmt)
{
    struct NwpTime	*timep;
    char		str[1024];
    int			r;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_optional(timep);
    r = strftime(str, 1023, STR2CSTR(fmt), &(timep->st));
    /* on errors (or empty output), strftime(3) returns 0. */
    str[r] = '\0';
    return rb_str_new2(str);
}

VALUE
nwpl_time_unix(VALUE obj)
{
    struct NwpTime	*time;
    Data_Get_Struct(obj, struct NwpTime, time);   
    return INT2NUM((time->min - UNIX_ERA) * MIN_PER_HOUR);
}

VALUE
nwpl_time_to_s(VALUE obj)
{
    char	buf[256], *ptr;
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    if (timep->min == -1) {
	ptr = "<NWPL::NwpTime OOB>";
    } else if (timep->min == 1) {
	ptr = "min1";
    } else {
        prepare_tm_essential(timep);
        snprintf(buf, 255, "%04d-%02d-%02dt%02d%02d",
	    timep->st.tm_year + 1900, timep->st.tm_mon + 1,
	    timep->st.tm_mday, timep->st.tm_hour, timep->st.tm_min);
        ptr = buf;
    }
    return rb_str_new2(ptr);
}

VALUE
nwpl_time_narrow(VALUE obj)
{
    char	buf[256];
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    snprintf(buf, 255, "%04d%02u%02u%02u%02u",
	timep->st.tm_year + 1900, timep->st.tm_mon + 1,
	timep->st.tm_mday, timep->st.tm_hour, timep->st.tm_min);
    return rb_str_new2(buf);
}

VALUE
nwpl_time_year(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    return INT2NUM(timep->st.tm_year + 1900);
}

VALUE
nwpl_time_mon(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    return INT2NUM(timep->st.tm_mon + 1);
}

VALUE
nwpl_time_mday(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    return INT2NUM(timep->st.tm_mday);
}

VALUE
nwpl_time_hour(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    return INT2NUM(timep->st.tm_hour);
}

VALUE
nwpl_time_min(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_essential(timep);
    return INT2NUM(timep->st.tm_min);
}

VALUE
nwpl_time_wday(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_optional(timep);
    return INT2NUM(timep->st.tm_wday);
}

VALUE
nwpl_time_yday(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    prepare_tm_optional(timep);
    return INT2NUM(timep->st.tm_yday);
}

VALUE
nwpl_time_undefp(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    return (timep->min == 1 || timep->min == -1) ? Qtrue : Qfalse;
}

VALUE
nwpl_time_oobp(VALUE obj)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    return (timep->min == -1) ? Qtrue : Qfalse;
}

static inline VALUE
new_time_pm(unsigned long pentad, long min, VALUE klass)
{
    struct NwpTime *tp;
    tp = ALLOC(struct NwpTime);
    tp->st.tm_min = min;
    tp->st.tm_hour = 0;
    tp->st.tm_mday = 1 + (pentad % 6) * 5;
    tp->st.tm_mon = (pentad / 6) % 12;
    tp->st.tm_year = (pentad / 6) / 12 - 4900;
    tp->st.tm_isdst = -2;
    tp->min = nwpl_ymdhm2seq(tp->st.tm_year + 1900, tp->st.tm_mon + 1,
	tp->st.tm_mday, tp->st.tm_hour, tp->st.tm_min);
    DEBUG(("%04d-%02d-%02d + %d min -> %d\n",
        tp->st.tm_year + 1900, tp->st.tm_mon + 1,
	tp->st.tm_mday, tp->st.tm_min, tp->min));
    return Data_Wrap_Struct(klass, 0, nwpl_time_free, tp);
}

static inline VALUE
add_time_tspan(struct NwpTime *tp, struct NwpTimeSpan *sp, int sign)
{
    unsigned long	pentad;
    long		remain;
    /* split date -> pentad and minutes */
    prepare_tm_essential(tp);
    pentad = (tp->st.tm_mday - 1) / 5;
    if (pentad > 5) pentad = 5;
    DEBUG(("#a tp=%02d/%02d pentad=%d sp=%d:%d\n",
	tp->st.tm_mon, tp->st.tm_mday, pentad,
	sign * sp->pentad, sign * sp->min));
    DEBUG(("#b %d %d %d\n", tp->st.tm_mday - 1 - pentad * 5,
	tp->st.tm_hour * MIN_PER_HOUR, tp->st.tm_min));
    remain = (tp->st.tm_mday - 1 - pentad * 5) * MIN_PER_DAY
	+ tp->st.tm_hour * MIN_PER_HOUR + tp->st.tm_min;
    pentad += (tp->st.tm_year + 4900) * 72 + (tp->st.tm_mon * 6);
    DEBUG(("#c pentad=%d remain=%d\n", pentad, remain));
    /* do add */
    pentad += sign * sp->pentad;
    remain += sign * sp->min;
    DEBUG(("#d pentad=%d remain=%d\n", pentad, remain));
    /* back to date */
    return new_time_pm(pentad, remain, nwpl_cNwpTime);
}

VALUE
nwpl_tspan_add(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return NEW_TSPAN(tspan->min + NUM2LONG(other), tspan->pentad,
		CLASS_OF(obj));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	return NEW_TSPAN(tspan->min + ospan->min,
	    tspan->pentad + ospan->pentad,
	    CLASS_OF(obj));
    } else if (CLASS_OF(other) == nwpl_cNwpTime) {
	struct NwpTime		*otime;
	Data_Get_Struct(other, struct NwpTime, otime);
	return add_time_tspan(otime, tspan, +1);
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

VALUE
nwpl_tspan_sub(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return NEW_TSPAN(tspan->min - NUM2LONG(other), tspan->pentad,
		CLASS_OF(obj));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	return NEW_TSPAN(tspan->min - ospan->min,
	    tspan->pentad - ospan->pentad,
	    CLASS_OF(obj));
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

static inline VALUE
int2cmp(int diff)
{
	if (diff > 0) return INT2FIX(1);
	if (diff < 0) return INT2FIX(-1);
	return INT2FIX(0);
}

VALUE
nwpl_tspan_cmp(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return int2cmp(tspan->min - NUM2LONG(other));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	if (tspan->min == ospan->min) {
	    return int2cmp(tspan->pentad - ospan->pentad);
	} else if (tspan->pentad == ospan->pentad) {
	    return int2cmp(tspan->min - ospan->min);
	} else {
	    /* assuming a pentad have exactly 5 days! */
	    DEBUG(("#span<=> %d %d %d\n", tspan->pentad * 7200 + tspan->min,
		ospan->pentad * 7200 + ospan->min,
	        int2cmp((tspan->pentad * 7200 + tspan->min)
		- (ospan->pentad * 7200 + ospan->min)) ));
	    return int2cmp((tspan->pentad * 7200 + tspan->min)
		- (ospan->pentad * 7200 + ospan->min));
	}
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

#define INT2BOOL(i)	((i) ? Qtrue : Qfalse)

VALUE
nwpl_tspan_eq(VALUE obj, VALUE other)
{
    struct NwpTimeSpan	*tspan;
    Data_Get_Struct(obj, struct NwpTimeSpan, tspan);   
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return INT2BOOL(tspan->min == NUM2LONG(other));
        default:
	    return Qfalse;
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	return INT2BOOL((tspan->min == ospan->min) &&
	    (tspan->pentad == ospan->pentad));
    } else {
	return Qfalse;
    }
}

VALUE
nwpl_time_add(VALUE obj, VALUE other)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return NEW_TIME(timep->min + NUM2LONG(other), CLASS_OF(obj));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	return add_time_tspan(timep, ospan, +1);
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

VALUE
nwpl_time_sub(VALUE obj, VALUE other)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return NEW_TIME(timep->min - NUM2LONG(other), CLASS_OF(obj));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTimeSpan) {
	struct NwpTimeSpan	*ospan;
	Data_Get_Struct(other, struct NwpTimeSpan, ospan);
	return add_time_tspan(timep, ospan, -1);
    } else if (CLASS_OF(other) == nwpl_cNwpTime) {
	struct NwpTime		*otime;
	Data_Get_Struct(other, struct NwpTime, otime);
	return NEW_TSPAN(timep->min - otime->min, 0, nwpl_cNwpTimeSpan);
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

VALUE
nwpl_time_cmp(VALUE obj, VALUE other)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return int2cmp(timep->min - NUM2LONG(other));
        default:
	    rb_raise(rb_eTypeError, "bad type");
    }
    if (CLASS_OF(other) == nwpl_cNwpTime) {
	struct NwpTime		*otime;
	Data_Get_Struct(other, struct NwpTime, otime);
	return int2cmp(timep->min - otime->min);
    } else {
	rb_raise(rb_eTypeError, "bad object class");
    }
}

VALUE
nwpl_time_eq(VALUE obj, VALUE other)
{
    struct NwpTime	*timep;
    Data_Get_Struct(obj, struct NwpTime, timep);
    switch (TYPE(other)) {
	case T_DATA:
	    break;
        case T_FLOAT:
        case T_FIXNUM:
        case T_BIGNUM:
	    return INT2BOOL(timep->min == NUM2LONG(other));
        default:
	    return Qfalse;
    }
    if (CLASS_OF(other) == nwpl_cNwpTime) {
	struct NwpTime		*otime;
	Data_Get_Struct(other, struct NwpTime, otime);
	return INT2BOOL(timep->min == otime->min);
    } else if (CLASS_OF(other) == rb_cTime) {
	unsigned long	s;
	s = NUM2ULONG(rb_funcall(other, rb_intern("to_i"), 0)) / MIN_PER_HOUR
		+ UNIX_ERA;
	return INT2BOOL(timep->min == s);
    } else {
        return Qfalse;
    }
}

void
Init_nwplraw()
{
    mNwpl = LookupModule("NWPL");

#define TIMESPAN nwpl_cNwpTimeSpan
    TIMESPAN
	= rb_define_class_under(mNwpl, "NwpTimeSpan", rb_cObject);
    rb_define_singleton_method(TIMESPAN, "new", nwpl_tspan_new, -1);
    rb_define_method(TIMESPAN, "min", nwpl_tspan_min, 0);
    rb_define_method(TIMESPAN, "mon", nwpl_tspan_mon, 0);
    rb_define_method(TIMESPAN, "pentad", nwpl_tspan_pentad, 0);
    rb_define_method(TIMESPAN, "to_s", nwpl_tspan_to_s, 0);
    rb_define_method(TIMESPAN, "*", nwpl_tspan_mul, 1);
    rb_define_method(TIMESPAN, "/", nwpl_tspan_div, 1);
    rb_define_method(TIMESPAN, "+", nwpl_tspan_add, 1);
    rb_define_method(TIMESPAN, "-", nwpl_tspan_sub, 1);
    rb_define_method(TIMESPAN, "<=>", nwpl_tspan_cmp, 1);
    rb_define_method(TIMESPAN, "==", nwpl_tspan_eq, 1);
    rb_define_method(TIMESPAN, "-@", nwpl_tspan_neg, 0);

    rb_define_const(TIMESPAN, "MIN", NEW_TSPAN(1, 0, TIMESPAN));
    rb_define_const(TIMESPAN, "HOUR", NEW_TSPAN(MIN_PER_HOUR, 0, TIMESPAN));
    rb_define_const(TIMESPAN, "DAY", NEW_TSPAN(MIN_PER_DAY, 0, TIMESPAN));
    rb_define_const(TIMESPAN, "WEEK", NEW_TSPAN(MIN_PER_WEEK, 0, TIMESPAN));
    rb_define_const(TIMESPAN, "PENT", NEW_TSPAN(0, 1, TIMESPAN));
    rb_define_const(TIMESPAN, "MON", NEW_TSPAN(0, 6, TIMESPAN));
    rb_define_const(TIMESPAN, "YEAR", NEW_TSPAN(0, 6 * 12, TIMESPAN));

    nwpl_cNwpTime
	= rb_define_class_under(mNwpl, "NwpTime", rb_cObject);
    rb_include_module(nwpl_cNwpTime, rb_mComparable);
    rb_define_singleton_method(nwpl_cNwpTime, "new", nwpl_time_new, -1);
    rb_define_singleton_method(nwpl_cNwpTime, "gettime", nwpl_time_gettime, -1);
    rb_define_method(nwpl_cNwpTime, "to_i", nwpl_time_to_i, 0);
    rb_define_const(nwpl_cNwpTime, "UNDEF", INT2FIX(1));
    rb_define_const(nwpl_cNwpTime, "UNIX_ERA", INT2FIX(UNIX_ERA));
    rb_define_method(nwpl_cNwpTime, "undef?", nwpl_time_undefp, 0);
    rb_define_method(nwpl_cNwpTime, "oob?", nwpl_time_oobp, 0);
    rb_define_method(nwpl_cNwpTime, "unix", nwpl_time_unix, 0);
    rb_define_method(nwpl_cNwpTime, "time", nwpl_time_time, 0);
    rb_define_method(nwpl_cNwpTime, "strftime", nwpl_time_strftime, 1);
    rb_define_method(nwpl_cNwpTime, "year", nwpl_time_year, 0);
    rb_define_method(nwpl_cNwpTime, "mon", nwpl_time_mon, 0);
    rb_define_method(nwpl_cNwpTime, "mday", nwpl_time_mday, 0);
    rb_define_method(nwpl_cNwpTime, "wday", nwpl_time_wday, 0);
    rb_define_method(nwpl_cNwpTime, "yday", nwpl_time_yday, 0);
    rb_define_method(nwpl_cNwpTime, "hour", nwpl_time_hour, 0);
    rb_define_method(nwpl_cNwpTime, "min", nwpl_time_min, 0);
    rb_define_method(nwpl_cNwpTime, "to_s", nwpl_time_to_s, 0);
    rb_define_method(nwpl_cNwpTime, "narrow", nwpl_time_narrow, 0);
    rb_define_method(nwpl_cNwpTime, "<=>", nwpl_time_cmp, 1);
    rb_define_method(nwpl_cNwpTime, "==", nwpl_time_eq, 1);
    rb_define_method(nwpl_cNwpTime, "-", nwpl_time_sub, 1);
    rb_define_method(nwpl_cNwpTime, "+", nwpl_time_add, 1);
}
