require 'nusdas'

module Nusdas
    class NusDef
    	def replace(str)
	    raise unless /^(\w+)\.(\w+)\.(\w+)$/ =~ nustype
	    model = $1[0..3]
	    co2d = $1[4..5]
	    co3d = $1[6..7]
	    attr = $2[0..1]
	    time = $2[2..3]
	    name = $3
	    r = str.dup
	    r.gsub!(/_model/, model)
	    r.gsub!(/_2d/i, co2d)
	    r.gsub!(/_3d/i, co3d)
	    r.gsub!(/_space/, co2d + co3d)
	    r.gsub!(/_attribute/, attr)
	    r.gsub!(/_time/, time)
	    r.gsub!(/_name/, name)
	    r
	end
    	def path_replaced
	    replace("#{path}/#{filename}")
	end
    end
end

class PathChange

    def self::help
	print <<EOF
#{$0}: restructure NuSDaS datasets directory structure

usage:
	#{$0} [-dc] conversion [ conversion ... ]

conversion:
	[member=MMMM] nrd1 def1 nrd2 def2

options:
	-d --debug	DEBUG mode
	-c --clobber	removes or overwrites existing files

EOF
	exit 0
   end

    def initialize(nrd1, def1, nrd2, def2)
	@nrd1, @nrd2 = nrd1, nrd2
	dset1 = Nusdas::NusDef::new("#{nrd1}/nusdas_def/#{def1}.def", nrd1)
	dset2 = Nusdas::NusDef::new("#{nrd2}/nusdas_def/#{def2}.def", nrd2)
	raise "nustype mismatch (#{dset1.nustype} != #{dset2.nustype})" \
	    if dset1.nustype != dset2.nustype
	p [nrd1, def1, nrd2, def2] if $DEBUG
	@p1 = dset1.path_replaced.split(/\/+/).reject{|a| /^\.?$/ =~ a}
	@p2 = dset2.path_replaced
	@link = {}
	@mkdir = {}
    end

    BVM = Regexp::new('_((base|valid)time|member)')

    def search(nrd1, p1, p2)
	if p1.empty?
	    raise "undefined #{$&}: #{p2}" if BVM =~ p2
	    raise "duplicated link: #{nrd1}, #{@link[p2]} --> #{p2}" \
	        if @link[p2]
	    @link[p2] = nrd1
	    return
	end
	print "search #{nrd1}\n"
	if BVM =~ p1.first
	    pat = p1.first.dup
	    pat.gsub!(/\W/, '\\\&')
	    keyword = []
	    pat.gsub!(BVM) do |str|
		keyword.push str
		str == '_member' ? '(\w{4})' : '(\d{12})'
	    end
	    rpat = Regexp::new("^#{pat}$")
	    dir = Dir::open(nrd1)
	    files = dir.collect
	    p [nrd1, p1, rpat] if $DEBUG
	    for file in files
		match = rpat.match(file)
		next unless match
		p [file, keyword, match[1..-1]] if $DEBUG
		myp2 = p2.dup
		1.upto(match.size - 1) do |i|
		    myp2.sub!(keyword[i-1], match[i])
		end
	        search("#{nrd1}/#{file}", p1[1..-1], myp2)
	    end
	else
	    raise unless File::exist?(dn = "#{nrd1}/#{p1.first}")
	    search(dn, p1[1..-1], p2)
	end
    end

    def mkdir_parent to
        parent = to.sub(%r|/+[^/]+/*$|, '')
	print "mkdir #{to}\n"
	return if @mkdir[parent]
	return @mkdir[parent] = true if File::directory? parent
	mkdir_parent(parent)
	Dir::mkdir parent
	@mkdir[parent] = true
    end

    def set_writable to
	return unless $OPT_clobber
        parent = to.sub(%r|/+[^/]+/*$|, '')
	print "chmod u+w #{to}\n"
	omode = File::stat(parent).mode
	File::chmod(omode | 0200, parent)
    end

    def link(from, to)
        mkdir_parent to
	print "ln #{from} #{to}\n"
	begin
	    File::link(from, to)
        rescue Errno::EEXIST
	    STDERR.print "#{to}: existent target removed\n"
	    print " ... existent target removed\n"
	    File::unlink(to) if $OPT_clobber
	    File::link(from, to)
	rescue Errno::EACCES
	    set_writable(to)
	    File::link(from, to)
	rescue Errno::EXDEV
	    raise unless $OPT_clobber
	    STDERR.print "#{to}: ln fails, and will try copying\n"
	    print " ... ln fails, and will try copying\n"
	    require 'ftools'
	    File::copy(from, to)
	end
    end

    def kick
	search(@nrd1, @p1, @p2)
	for to, from in @link
	    link from, "#@nrd2/#{to}"
	end
    end

    def member=(str)
        return if str.nil?
        str = [str].pack('A4').gsub(/ /, '_')
	@p2.gsub!(/_member/, str)
    end

    def self::main
        require 'getopts'
	getopts('dwc', *%w(clobber debug verbose))
	$DEBUG = true if $OPT_d or $OPT_debug
	$VERBOSE = true if $OPT_w or $OPT_verbose
	$OPT_clobber = true if $OPT_c
	help if (ARGV.size == 0)
	member = nil
	while not ARGV.empty?
	    if /^member=/ =~ ARGV.first
		member = $'
		ARGV.shift
	    end
	    nrd1 = ARGV.shift
	    def1 = ARGV.shift
	    nrd2 = ARGV.shift
	    def2 = ARGV.shift
	    help if def2.nil?
	    conv = self.new(nrd1, def1, nrd2, def2)
	    conv.member = member if member
	    conv.kick
	end
    rescue => e
        STDERR.print e.message, "\n"
        print e.message, "\n"
        help
    end

end


PathChange.main
exit 0

