/* endian.c - byte order wrapper
 * Copyright (C) by Japan Meteorological Agency, 2001.  All rights reserved.
 * 2001-07-02 TOYODA Eizi
 */

static char rcsid[] =
    "$Header: /cvsroot/pandora/pnusdas/src/endian.c,v 1.15 2003/02/24 02:16:42 tabito Exp $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "nusdas_endian.h"
#include "config_endian.h"

#ifdef IS_LITTLE_ENDIAN

void swab2(char *ptr, size_t nmemb)
{
    size_t offset;
    for (offset = 0; offset < nmemb * 2; offset += 2) {
	char a, b;
	a = ptr[offset];
	b = ptr[offset + 1];
	ptr[offset] = b;
	ptr[offset + 1] = a;
    }
}

void swab4(char *ptr, size_t nmemb)
{
    size_t offset;
    for (offset = 0; offset < nmemb * 4; offset += 4) {
	char a, b, c, d;
	a = ptr[offset];
	b = ptr[offset + 1];
	c = ptr[offset + 2];
	d = ptr[offset + 3];
	ptr[offset] = d;
	ptr[offset + 1] = c;
	ptr[offset + 2] = b;
	ptr[offset + 3] = a;
    }
}

void swab8(char *ptr, size_t nmemb)
{
    size_t offset;
    char buf[8];
    int idx;
    for (offset = 0; offset < nmemb * 8; offset += 8) {
	for (idx = 0; idx < 8; idx++)
	    buf[idx] = ptr[offset + idx];
	for (idx = 0; idx < 8; idx++)
	    ptr[offset + idx] = buf[8 - 1 - idx];
    }
}

void swab16(char *ptr, size_t nmemb)
{
    size_t offset;
    char buf[16];
    int idx;
    for (offset = 0; offset < nmemb * 16; offset += 16) {
	for (idx = 0; idx < 16; idx++)
	    buf[idx] = ptr[offset + idx];
	for (idx = 0; idx < 16; idx++)
	    ptr[offset + idx] = buf[16 - 1 - idx];
    }
}

# define SWAB2(ptr, nmemb) swab2(ptr, nmemb)
# define SWAB4(ptr, nmemb) swab4(ptr, nmemb)
# define SWAB8(ptr, nmemb) swab8(ptr, nmemb)
# define SWAB16(ptr, nmemb) swab16(ptr, nmemb)

#else				/* IS_LITTLE_ENDIAN */

# define SWAB2(ptr, nmemb) 1
# define SWAB4(ptr, nmemb) 1
# define SWAB8(ptr, nmemb) 1
# define SWAB16(ptr, nmemb) 1

#endif				/* IS_LITTLE_ENDIAN */

    size_t
fread_native(void *ptr, size_t size, size_t nmemb, FILE * stream)
{
    size_t r;
    r = fread(ptr, size, nmemb, stream);
#ifdef IS_LITTLE_ENDIAN
    switch (size) {
    case 1:
	;
	break;
    case 2:
	SWAB2(ptr, nmemb);
	break;
    case 4:
	SWAB4(ptr, nmemb);
	break;
    case 8:
	SWAB8(ptr, nmemb);
	break;
    case 16:
	SWAB16(ptr, nmemb);
	break;
    default:
	errprintf( "fread_native(%p, invalid:%d, %d, %p)\n",
	    ptr, size, nmemb, stream);
	return -1;
    }
#endif
    return r;
}

size_t fread_fmt(void *ptr, char *fmt, FILE * stream)
{
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t sum, r;

    sum = 0;
    p = ptr;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
	if (conv == 'q' || conv == 'Q') {
	    r = fread(p, 16, count, stream);
	    SWAB16(p, count);
	    p += r * 16;
	} else if (conv == 'd' || conv == 'D' || conv == 'l'
		   || conv == 'L') {
	    r = fread(p, 8, count, stream);
	    SWAB8(p, count);
	    p += r * 8;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
	    r = fread(p, 4, count, stream);
	    SWAB4(p, count);
	    p += r * 4;
	} else if (conv == 'h' || conv == 'H') {
	    r = fread(p, 2, count, stream);
	    SWAB2(p, count);
	    p += r * 2;
	} else {
	    r = fread(p, 1, count, stream);
	    p += r;
	}
    }
    return p - (char *) ptr;
}

    size_t
fwrite_native(void *ptr, size_t size, size_t nmemb, FILE * stream)
{
    size_t r;
#ifndef IS_LITTLE_ENDIAN
    r = fwrite(ptr, size, nmemb, stream);
#else
# define BLOCK 4096
    size_t	nelem, n, ret;
    char	buffer[BLOCK];

    switch (size) {
    case 1:
	r = fwrite(ptr, size, nmemb, stream);
	break;
    case 2:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 2;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 2);
		SWAB2(buffer, n);
		ret = fwrite(buffer, 2, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 2;
		nelem -= n;
		r += ret;
	}
	break;
    case 4:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 4;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 4);
		SWAB4(buffer, n);
		ret = fwrite(buffer, 4, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 4;
		nelem -= n;
		r += ret;
	}
	break;
    case 8:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 8;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 8);
		SWAB8(buffer, n);
		ret = fwrite(buffer, 8, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 8;
		nelem -= n;
		r += ret;
	}
	break;
    case 16:
	nelem = nmemb;
	r = 0;
	while (nelem > 0) {
		n = BLOCK / 16;
		if (nelem < n) n = nelem;
		memcpy(buffer, ptr, n * 16);
		SWAB16(buffer, n);
		ret = fwrite(buffer, 16, n, stream);
		if (ret < n) break;
		ptr = (char *)ptr + n * 16;
		nelem -= n;
		r += ret;
	}
	break;
    default:
	errprintf( "fwrite_native(%p, invalid:%d, %d, %p)\n",
		ptr, size, nmemb, stream);
	return -1;
    }
#endif
    return r;
}

    size_t
fwrite_fmt(void *ptr, char *fmt, FILE * stream)
{
    char *p, *f, *f2;
    int conv;
    unsigned long count;
    size_t sum, r;

    sum = 0;
    p = ptr;
    for (f = fmt; *f; f++) {
	count = strtoul(f, &f2, 0);
	if (f2 == f) {
	    count = 1;
	    conv = *f2;
	} else if (f2 == NULL) {
	    conv = 'b';
	    /* this causes loop-break after f++ */
	    f = "*";
	} else {
	    conv = *f2;
	    f = f2;
	}
	if (conv == 'q' || conv == 'Q') {
	    r = fwrite_native(p, 16, count, stream);
	    p += r * 16;
	} else if (conv == 'd' || conv == 'D' || conv == 'l'
		   || conv == 'L') {
	    r = fwrite_native(p, 8, count, stream);
	    p += r * 8;
	} else if (conv == 'f' || conv == 'F' || conv == 'i'
		   || conv == 'I') {
	    r = fwrite_native(p, 4, count, stream);
	    p += r * 4;
	} else if (conv == 'h' || conv == 'H') {
	    r = fwrite_native(p, 2, count, stream);
	    p += r * 2;
	} else {
	    r = fwrite(p, 1, count, stream);
	    p += r;
	}
    }
    return p - (char *) ptr;
}


/*
 * ^ϊ[`
 */

/*
 * big endian z 4 oCg long l
 */
	long
btol(const unsigned char *ptr)
{
	long	r;
	r = ptr[0];
	r <<= 8;
	r |= ptr[1];
	r <<= 8;
	r |= ptr[2];
	r <<= 8;
	r |= ptr[3];
	return r;
}

/*
2003.2.24 tabito added
byte order ϊ[`
*/

void 
endian_swab2(char *ptr, size_t nmemb)
{
    SWAB2(ptr,nmemb);
    return;
}

void 
endian_swab4(char *ptr, size_t nmemb)
{
    SWAB4(ptr,nmemb);
    return;
}

void 
endian_swab8(char *ptr, size_t nmemb)
{
    SWAB8(ptr,nmemb);
    return;
}

void 
endian_swab16(char *ptr, size_t nmemb)
{
    SWAB16(ptr,nmemb);
    return;
}
